
; *******************************************************************
;
;        Software License Agreement
;
; The software supplied herewith by Microchip Technology Incorporated 
; (the Company) for its PICmicro Microcontroller is intended and 
; supplied to you, the Companys customer, for use solely and 
; exclusively on Microchip PICmicro Microcontroller products. The 
; software is owned by the Company and/or its supplier, and is 
; protected under applicable copyright laws. All rights are reserved. 
; Any use in violation of the foregoing restrictions may subject the 
; user to criminal sanctions under applicable laws, as well as to 
; civil liability for the breach of the terms and conditions of this 
; license.
;
; THIS SOFTWARE IS PROVIDED IN AN AS IS CONDITION. NO WARRANTIES, 
; WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED 
; TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
; PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE COMPANY SHALL NOT, 
; IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR 
; CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
;
; *******************************************************************

; *******************************************************************
	Title	"RIJNDAEL Decryption Implementation Software"
; *******************************************************************

; *******************************************************************
; ***								  ***
; ***	Author:	Caio Gubel 					  ***
; ***		Applications Engineer				  ***
; ***		Microchip Technology Inc.			  ***
; ***								  ***
; ***  Program:	aes_decr.asm					  ***
; ***		This routine takes 16 bytes of AES/Rijndael 	  ***
; *** 		algorithm encrypted data and decrypts it 	  ***
; ***		using the AES/Rijndael algorithm. 		  ***
; ***								  ***
; *** Last Rev:	April 27, 2001					  ***
; ***		Ver 1.00					  ***
; ***								  ***
; *******************************************************************



; *******************************************************************	   

	list	P=PIC16C622A
	#include p16c622A.inc
	#include TABLES.inc		; Holds S_TABLE & Si_TABLE Data 
					;    Starts at Location 0x0600

; *******************************************************************


;*********************************************************************
;
	__CONFIG   _CP_OFF & _WDT_ON & _BODEN_OFF & _PWRTE_ON & _RC_OSC



;****************** Constants ************************

INIT_RAM0	equ	0x20		; Starting point of RAM in segment 0
ROUNDS		equ	0x0A		; number of rounds necessary to cipher/decipher


;****************** Variables ************************

block		equ	0x20		; block vector ( 16 bytes*8 => 128 bits) 
key		equ	0x30		; key vector ( 16 bytes*8 => 128 bits)

aux		equ	0x40		; general purpose var
aux1		equ	0x41		; 
aux2		equ	0x42
aux3		equ	0x43

round_counter	equ	0x44		; encryption/decryption rounds counter
rcon		equ	0x45		;

d0		equ	0x4A
d1		equ	0x4B
d2		equ	0x4C

		ORG 	0x0000

start:
	goto	main			;

		ORG	0x0004

int_handler:
	

; ***************** END Interrupt Handler ***********************


		ORG	0x0010


;********************** MAIN Routine ****************************
; This routine deciphers a given text.
; The  init_decryption_key routine was put here to show how to 
; create a decryption key from the initial key
;
main:
	call	init_cpu		; start CPU registers
	call	set_test_block		; loads the ciphered text to be deciphered
	call	set_test_key		; load a second time the key used to cipher
	call	init_decryption_key	; create the initial decryption key
	call	decrypt			; decipher the block back to plain text	
	goto	$



; ***************** INIT_CPU ************************************
; This procedure cleans the RAM in segment 0 
; Input: None
; Output: None
; ***************************************************************

init_cpu:
	movlw   INIT_RAM0                       ;points to the start of the RAM
	movwf   FSR
loop_cram0
	clrf    INDF				;clear RAM location pointed by FSR
	incf    FSR,f                           ;increment pointer
	btfss   FSR,7                           ;while FSR < 0x80 loop
	goto    loop_cram0                      ;
	return					




; ********************* SET_TEST_BLOCK **************************
; This procedure sets the test ciphered text to be deciphered
; Input: 	
; Outup: 	block[16]={B0,B1,...}
; Globals: 	block,aux,FSR
; OBS:		Generates block vector for test; 
; ***************************************************************

set_test_block:
	movlw	0xb9
	movwf	block+0x0
	movlw	0x70
	movwf	block+0x1
	movlw	0xdf
	movwf	block+0x2
	movlw	0xbe
	movwf	block+0x3
	movlw	0x40
	movwf	block+0x4
	movlw	0x69
	movwf	block+0x5
	movlw	0x8a
	movwf	block+0x6
	movlw	0xf1
	movwf	block+0x7
	movlw	0x63
	movwf	block+0x8
	movlw	0x8f
	movwf	block+0x9
	movlw	0xe3
	movwf	block+0xA
	movlw	0x8b
	movwf	block+0xB
	movlw	0xd3
	movwf	block+0xC
	movlw	0xdf
	movwf	block+0xD
	movlw	0x3b
	movwf	block+0xE
	movlw	0x2f
	movwf	block+0xF
	return 

; ********************* SET_TEST_KEY **************************
; This procedure sets the key used to cipher
; Input: 	None
; Outup: 	key[16]
; Globals: 	key[16],FSR,aux
; ***************************************************************

set_test_key:

	movlw	0x0
	movwf	key+0x0
	movlw	0x0
	movwf	key+0x1
	movlw	0x0
	movwf	key+0x2
	movlw	0x0
	movwf	key+0x3
	movlw	0x0
	movwf	key+0x4
	movlw	0x0
	movwf	key+0x5
	movlw	0x0
	movwf	key+0x6
	movlw	0x0
	movwf	key+0x7
	movlw	0x00
	movwf	key+0x8
	movlw	0x0
	movwf	key+0x9
	movlw	0x0
	movwf	key+0xA
	movlw	0x0
	movwf	key+0xB
	movlw	0x0
	movwf	key+0xC
	movlw	0x0
	movwf	key+0xD
	movlw	0x0
	movwf	key+0xE
	movlw	0x0
	movwf	key+0xF
	return 

; ********************* KEY_ADDITION ****************************
; This procedure makes: block ^= key ( ^= means block = block XOR key)
; Input: 	None
; Outup: 	
; Globals: 	block, key
; OBS: 		Implemented without a loop to increase speed
; ***************************************************************

key_addition:
	movf	key+0x0,w			; block[0] ^= key[0]; 
	xorwf	block+0x0,f

	movf	key+0x1,w			; block[1] ^= key[1]; 
	xorwf	block+0x1,f

	movf	key+0x2,w			; block[2] ^= key[2]; 
	xorwf	block+0x2,f

	movf	key+0x3,w			; block[3] ^= key[3]; 
	xorwf	block+0x3,f

	movf	key+0x4,w			; block[4] ^= key[4]; 
	xorwf	block+0x4,f

	movf	key+0x5,w			; block[5] ^= key[5]; 
	xorwf	block+0x5,f

	movf	key+0x6,w			; block[6] ^= key[6]; 
	xorwf	block+0x6,f

	movf	key+0x7,w			; block[7] ^= key[7]; 
	xorwf	block+0x7,f

	movf	key+0x8,w			; block[8] ^= key[8]; 
	xorwf	block+0x8,f

	movf	key+0x9,w			; block[9] ^= key[9]; 
	xorwf	block+0x9,f

	movf	key+0x0A,w			; block[10] ^= key[10]; 
	xorwf	block+0x0A,f

	movf	key+0x0B,w			; block[11] ^= key[11]; 
	xorwf	block+0x0B,f

	movf	key+0x0C,w			; block[12] ^= key[12]; 
	xorwf	block+0x0C,f

	movf	key+0x0D,w			; block[13] ^= key[13]; 
	xorwf	block+0x0D,f

	movf	key+0x0E,w			; block[14] ^= key[14]; 
	xorwf	block+0x0E,f

	movf	key+0x0F,w			; block[15] ^= key[15]; 
	xorwf	block+0x0F,f
	return



; *********************** SUBSTITUTION_Si *************************
; This procedure makes: block[i] = Si_box[block[i]].
; This is the inverse mapping because:
;
;	b = S_box[a] and
;	a = Si_box[b]
;			
; Input: 	
; Outup: 	
; Globals: aux,block,FSR,PCLATH
; *****************************************************************

substitution_Si: 			; implements the inverse substitution
	movlw	HIGH Si_table		; selects segment 
	movwf	PCLATH			; 
					; it works like: for i(0....15)	
put_si_table_b0:
	movlw	D'82'			; if (block[i]==0) block[]=82D
	movf	block+0x0,f
	btfsc	STATUS,Z
	goto	putsi_b0
	decf	block+0x0,w
	call	Si_table		; else block[i] = Si_table[block[i]-1]
putsi_b0:
	movwf	block+0x0

put_si_table_b1:
	movlw	D'82'			; 
	movf	block+0x1,f
	btfsc	STATUS,Z
	goto	putsi_b1
	decf	block+0x1,w
	call	Si_table		; 
putsi_b1:
	movwf	block+0x1

put_si_table_b2:
	movlw	D'82'			; 
	movf	block+0x2,f
	btfsc	STATUS,Z
	goto	putsi_b2
	decf	block+0x2,w
	call	Si_table		;
putsi_b2:
	movwf	block+0x2

put_si_table_b3:
	movlw	D'82'			;
	movf	block+0x3,f
	btfsc	STATUS,Z
	goto	putsi_b3
	decf	block+0x3,w
	call	Si_table		;
putsi_b3:
	movwf	block+0x3

put_si_table_b4:
	movlw	D'82'			;
	movf	block+0x4,f
	btfsc	STATUS,Z
	goto	putsi_b4
	decf	block+0x4,w
	call	Si_table		;
putsi_b4:
	movwf	block+0x4

put_si_table_b5:
	movlw	D'82'			;
	movf	block+0x5,f
	btfsc	STATUS,Z
	goto	putsi_b5
	decf	block+0x5,w
	call	Si_table		;
putsi_b5:
	movwf	block+0x5

put_si_table_b6:
	movlw	D'82'			;
	movf	block+0x6,f
	btfsc	STATUS,Z
	goto	putsi_b6
	decf	block+0x6,w
	call	Si_table		;
putsi_b6:
	movwf	block+0x6

put_si_table_b7:
	movlw	D'82'			;
	movf	block+0x7,f
	btfsc	STATUS,Z
	goto	putsi_b7
	decf	block+0x7,w
	call	Si_table		;
putsi_b7:
	movwf	block+0x7

put_si_table_b8:
	movlw	D'82'			;
	movf	block+0x8,f
	btfsc	STATUS,Z
	goto	putsi_b8
	decf	block+0x8,w
	call	Si_table		;
putsi_b8:
	movwf	block+0x8

put_si_table_b9:
	movlw	D'82'			;
	movf	block+0x9,f
	btfsc	STATUS,Z
	goto	putsi_b9
	decf	block+0x9,w
	call	Si_table		;
putsi_b9:
	movwf	block+0x9

put_si_table_bA:
	movlw	D'82'			;
	movf	block+0xA,f
	btfsc	STATUS,Z
	goto	putsi_bA
	decf	block+0xA,w
	call	Si_table		;
putsi_bA:
	movwf	block+0xA

put_si_table_bB:
	movlw	D'82'			;
	movf	block+0xB,f
	btfsc	STATUS,Z
	goto	putsi_bB
	decf	block+0xB,w
	call	Si_table		;
putsi_bB:
	movwf	block+0xB

put_si_table_bC:
	movlw	D'82'			;
	movf	block+0xC,f
	btfsc	STATUS,Z
	goto	putsi_bC
	decf	block+0xC,w
	call	Si_table		;
putsi_bC:
	movwf	block+0xC

put_si_table_bD:
	movlw	D'82'			;
	movf	block+0xD,f
	btfsc	STATUS,Z
	goto	putsi_bD
	decf	block+0xD,w
	call	Si_table		;
putsi_bD:
	movwf	block+0xD

put_si_table_bE:
	movlw	D'82'			;
	movf	block+0xE,f
	btfsc	STATUS,Z
	goto	putsi_bE
	decf	block+0xE,w
	call	Si_table		;
putsi_bE:
	movwf	block+0xE

put_si_table_bF:
	movlw	D'82'			;
	movf	block+0xF,f
	btfsc	STATUS,Z
	goto	putsi_bF
	decf	block+0xF,w
	call	Si_table		;
putsi_bF:
	movwf	block+0xF
	return


; *********************** DEC_SHIFT_ROW *******************************
; These procedures shift rows (1,2 and 3 but NOT 0) a variable 
; amount on the direction (decryption)
; Input: 	 
; Outup: 	shifted block
; Globals: 	aux, block, row_vector
; OBS: 
; 	As defined in the AES Proposal: Rijndael document,
;   	this procedure ciclically shifts rows 1,2 and 3 by offsets
;	that for 128 bits blocks and key are defined as:
;	Cipher:		row1 --> 1, row2 -->2 and row3 -->3
;	Decipher:	row1 --> 3, row2 -->2 and row3 -->1.
;
;	It is also equivalent to the folowing reference 
;	C code designed by Paulo Barreto and Vincent Rijmen:
;	** parameter "d" gives the direction: 	0 = cipher
;						1 = decipher 
;	
;	static word8 shifts[3][4][2] = {0, 0,
;					1, 3,
;					2, 2,
;					3, 1,
;
;					0, 0,
;					1, 5,
;					2, 4,
;					3, 3,
;
;					0, 0,
;					1, 7,
;					3, 5,
;					4, 4}; 
;
;	void ShiftRow(word8 a[4][MAXBC], word8 d, word8 BC) 
;	{
;	/* Row 0 remains unchanged * The other three rows are shifted a variable amount*/
;	 word8 tmp[MAXBC];
;	 int i, j;
;
;	 for(i = 1; i < 4; i++) 
;	 {
;	  for(j = 0; j < BC; j++) tmp[j] = a[i][(j + shifts[SC][i][d]) % BC];
;         for(j = 0; j < BC; j++) a[i][j] = tmp[j];
;	 }
;       }
;
;	Obs: BC = 4, SC=0 
;	d=0 -> encription
;	d=1 -> encription
;
; ***************************************************************************************
;

dec_shift_row:

; Rotates right row 1 one position 

	movf	block+0xD,w
	movwf	aux
	movf	block+0x9,w
	movwf	block+0xD
	movf	block+0x5,w
	movwf	block+0x9
	movf	block+0x1,w
	movwf	block+0x5
	movf	aux,w
	movwf	block+0x1
	
; Rotates right row 2 two positions 

	movf	block+0xA,w
	movwf	aux
	movf	block+0xE,w
	movwf	aux1
	movf	block+0x2,w
	movwf	block+0xA
	movf	block+0x6,w
	movwf	block+0xE
	movf	aux,w
	movwf	block+0x2
	movf	aux1,w
	movwf	block+0x6

; Rotates right row 3 one positions 

	movf	block+0x3,w
	movwf	aux
	movf	block+0x7,w
	movwf	block+0x3
	movf	block+0xB,w
	movwf	block+0x7
	movf	block+0xF,w
	movwf	block+0xB
	movf	aux,w
	movwf	block+0xF
	return
	
; *********************** INV_MIX_COLUMN ************************
; This procedure calculates the inverse mix column of the state
; (block with partial results); it is equivalent to the 
; multiplication of the block taken in rows of an 4x4 matrix by 
; the following matrix (over GF2)
; 	
;	0xE 0xB 0xD 0x9	
;	0xB 0xD 0x9 0xE	
;	0xD 0x9 0xE 0xB	
;	0x9 0xE 0xB 0xD	
;	
; The equivalent C code designed by Paulo Barreto and Vincent Rijmen:
;
; void InvMixColumn(word8 a[4][MAXBC], word8 BC) 
;  {/* Mix the four bytes of every column in a linear way
;    * This is the opposite operation of Mixcolumn */
;   word8 b[4][MAXBC];
;   int i, j;
;
;   for(j = 0; j < BC; j++)
;    for(i = 0; i < 4; i++);
;      b[i][j] = mul(0xe,a[i][j])
;      		 ^ mul(0xb,a[(i + 1) % 4][j])
;		 ^ mul(0xd,a[(i + 2) % 4][j])
;		 ^ mul(0x9,a[(i + 3) % 4][j]);
;   for(i = 0; i < 4; i++)
;    for(j = 0; j < BC; j++) a[i][j] = b[i][j];
;  }
;
;	Obs: BC=4
;
;	
; Input: 	block
; Outup: 	block
; Globals: 	aux,aux1,aux2,aux3
;
; OBS:This implementation uses the following optimized equivalent 
;     structure: (as seen in Geoffrey Keating's implementation)
;
; 	aux  = block[i+0] ^ block[i+0x1] ^ block[i+0x2] ^ block[i+0x3];
;	aux1 = xtime(block[i+0x0] ^ block[i+0x2]);
;	aux2 = xtime(block[i+1] ^ block[i+0x3]);
;	aux3 = xtime( xtime( aux1 ^ aux2 ) ^ aux;
;	d0   = xtime(block[i+0x0]^block[i+0x1] ^ aux1);
;	d1   = xtime(block[i+0x1] ^ block[i+0x2] ^ aux2);
;	d2   = xtime(block[i+0x2] ^ block[i+0xC] ^ aux1);
;	Block[i+0x0]   ^= d0 ^ aux3;	
;	Block[i+0x1]   ^= d1 ^ aux3;
;	block[i+0x2]   ^= d2 ^ aux3;
;	Block[i+0x3]    = block[i+0x0] ^ block[i+1] ^ block[i+2] ^ aux;
;
; ***************************************************************

	
inv_mix_column:

; **** it begins by calculating inv_mix_column for the FIRST column ****

	movf	block+0x0,w		; aux = block+0x0 ^ block+0x1 ^ block+0x2 ^ block+0x3 		
	xorwf	block+0x1,w	
	xorwf	block+0x2,w
	xorwf	block+0x3,w
	movwf	aux

	movf	block+0x0,w		; aux1 = xtime(block+0x0 ^ block+0x2)
	xorwf	block+0x2,w
	movwf	aux1
	movlw	0x1B
	bcf	STATUS,C
	rlf	aux1,f
	btfsc	STATUS,C
	xorwf	aux1,f
	
	movf	block+0x1,w		; aux2 = xtime(block+0x1 ^ block+0x3)
	xorwf	block+0x3,w		
	movwf	aux2
	movlw	0x1B
	bcf	STATUS,C
	rlf	aux2,f
	btfsc	STATUS,C
	xorwf	aux2,f
	
	movf	aux1,w			; aux3 = xtime(xtime(block+0x1 ^ block+0x2)) ^ aux
	xorwf	aux2,w		
	movwf	aux3
	movlw	0x1B
	bcf	STATUS,C
	rlf	aux3,f
	btfsc	STATUS,C
	xorwf	aux3,f
	bcf	STATUS,C
	rlf	aux3,f
	btfsc	STATUS,C
	xorwf	aux3,f
	movf	aux,w
	xorwf	aux3,f
	
	movf	block+0x0,w		; d0 = xtime(block+0x0 ^ block+0x1 ^ aux1)
	xorwf	block+0x1,w
	xorwf	aux1,w
	movwf	d0
	movlw	0x1B
	bcf	STATUS,C
	rlf	d0,f
	btfsc	STATUS,C
	xorwf	d0,f

	movf	block+0x1,w		; d1 = xtime(block+0x1 ^ block+0x2 ^ aux2)
	xorwf	block+0x2,w
	xorwf	aux2,w
	movwf	d1
	movlw	0x1B
	bcf	STATUS,C
	rlf	d1,f
	btfsc	STATUS,C
	xorwf	d1,f

	
	movf	block+0x2,w		; d2 = xtime(block+0x2 ^ block+0x3 ^ aux1)
	xorwf	block+0x3,w
	xorwf	aux1,w
	movwf	d2
	movlw	0x1B
	bcf	STATUS,C
	rlf	d2,f
	btfsc	STATUS,C
	xorwf	d2,f


	movf	d0,w			; block+0 ^= d0 ^ aux3
	xorwf	aux3,w
	xorwf	block+0x0,f

	movf	d1,w			; block+1 ^= d1 ^ aux3
	xorwf	aux3,w
	xorwf	block+0x1,f
	
	movf	d2,w			; block+2 ^= d2 ^ aux3
	xorwf	aux3,w
	xorwf	block+0x2,f

	movf	block+0x0,w		; block+3 = block+0 ^ block+1 ^ block+2 ^ aux
	xorwf	block+0x1,w
	xorwf	block+0x2,w
	xorwf	aux,w
	movwf	block+0x3
	
; ******* inv_mix_column for the SECOND column *******
; It uses the same structure used in the first column

	movf	block+0x4,w		; 
	xorwf	block+0x5,w	
	xorwf	block+0x6,w
	xorwf	block+0x7,w
	movwf	aux

	movf	block+0x4,w		; 
	xorwf	block+0x6,w
	movwf	aux1
	movlw	0x1B
	bcf	STATUS,C
	rlf	aux1,f
	btfsc	STATUS,C
	xorwf	aux1,f
	
	movf	block+0x5,w		; 
	xorwf	block+0x7,w		
	movwf	aux2
	movlw	0x1B
	bcf	STATUS,C
	rlf	aux2,f
	btfsc	STATUS,C
	xorwf	aux2,f
	
	movf	aux1,w			;
	xorwf	aux2,w		
	movwf	aux3
	movlw	0x1B
	bcf	STATUS,C
	rlf	aux3,f
	btfsc	STATUS,C
	xorwf	aux3,f
	bcf	STATUS,C
	rlf	aux3,f
	btfsc	STATUS,C
	xorwf	aux3,f
	movf	aux,w
	xorwf	aux3,f
	
	movf	block+0x4,w		;
	xorwf	block+0x5,w
	xorwf	aux1,w
	movwf	d0
	movlw	0x1B
	bcf	STATUS,C
	rlf	d0,f
	btfsc	STATUS,C
	xorwf	d0,f

	movf	block+0x5,w		;
	xorwf	block+0x6,w
	xorwf	aux2,w
	movwf	d1
	movlw	0x1B
	bcf	STATUS,C
	rlf	d1,f
	btfsc	STATUS,C
	xorwf	d1,f

	
	movf	block+0x6,w		;
	xorwf	block+0x7,w
	xorwf	aux1,w
	movwf	d2
	movlw	0x1B
	bcf	STATUS,C
	rlf	d2,f
	btfsc	STATUS,C
	xorwf	d2,f


	movf	d0,w			;
	xorwf	aux3,w
	xorwf	block+0x4,f

	movf	d1,w			;
	xorwf	aux3,w
	xorwf	block+0x5,f
	
	movf	d2,w			;
	xorwf	aux3,w
	xorwf	block+0x6,f

	movf	block+0x4,w		;
	xorwf	block+0x5,w
	xorwf	block+0x6,w
	xorwf	aux,w
	movwf	block+0x7


; ******* inv_mix_column for the THIRD column *******
; It uses the same structure used in the first column

	movf	block+0x8,w	
	xorwf	block+0x9,w	
	xorwf	block+0xA,w
	xorwf	block+0xB,w
	movwf	aux

	movf	block+0x8,w		; 
	xorwf	block+0xA,w
	movwf	aux1
	movlw	0x1B
	bcf	STATUS,C
	rlf	aux1,f
	btfsc	STATUS,C
	xorwf	aux1,f
	
	movf	block+0x9,w		; 
	xorwf	block+0xB,w		
	movwf	aux2
	movlw	0x1B
	bcf	STATUS,C
	rlf	aux2,f
	btfsc	STATUS,C
	xorwf	aux2,f
	
	movf	aux1,w			;
	xorwf	aux2,w		
	movwf	aux3
	movlw	0x1B
	bcf	STATUS,C
	rlf	aux3,f
	btfsc	STATUS,C
	xorwf	aux3,f
	bcf	STATUS,C
	rlf	aux3,f
	btfsc	STATUS,C
	xorwf	aux3,f
	movf	aux,w
	xorwf	aux3,f
	
	movf	block+0x8,w		;
	xorwf	block+0x9,w
	xorwf	aux1,w
	movwf	d0
	movlw	0x1B
	bcf	STATUS,C
	rlf	d0,f
	btfsc	STATUS,C
	xorwf	d0,f

	movf	block+0x9,w		;
	xorwf	block+0xA,w
	xorwf	aux2,w
	movwf	d1
	movlw	0x1B
	bcf	STATUS,C
	rlf	d1,f
	btfsc	STATUS,C
	xorwf	d1,f

	
	movf	block+0xA,w		;
	xorwf	block+0xB,w
	xorwf	aux1,w
	movwf	d2
	movlw	0x1B
	bcf	STATUS,C
	rlf	d2,f
	btfsc	STATUS,C
	xorwf	d2,f


	movf	d0,w			;
	xorwf	aux3,w
	xorwf	block+0x8,f

	movf	d1,w			;
	xorwf	aux3,w
	xorwf	block+0x9,f
	
	movf	d2,w			;
	xorwf	aux3,w
	xorwf	block+0xA,f

	movf	block+0x8,w		;
	xorwf	block+0x9,w
	xorwf	block+0xA,w
	xorwf	aux,w
	movwf	block+0xB

; ******* inv_mix_column for the LAST (FOURTH) column *******
; It uses the same structure used in the first column


	movf	block+0xC,w		; 
	xorwf	block+0xD,w	
	xorwf	block+0xE,w
	xorwf	block+0xF,w
	movwf	aux

	movf	block+0xC,w		;
	xorwf	block+0xE,w
	movwf	aux1
	movlw	0x1B
	bcf	STATUS,C
	rlf	aux1,f
	btfsc	STATUS,C
	xorwf	aux1,f
	
	movf	block+0xD,w		;
	xorwf	block+0xF,w		
	movwf	aux2
	movlw	0x1B
	bcf	STATUS,C
	rlf	aux2,f
	btfsc	STATUS,C
	xorwf	aux2,f
	
	movf	aux1,w			;
	xorwf	aux2,w		
	movwf	aux3
	movlw	0x1B
	bcf	STATUS,C
	rlf	aux3,f
	btfsc	STATUS,C
	xorwf	aux3,f
	bcf	STATUS,C
	rlf	aux3,f
	btfsc	STATUS,C
	xorwf	aux3,f
	movf	aux,w
	xorwf	aux3,f
	
	movf	block+0xC,w		;
	xorwf	block+0xD,w
	xorwf	aux1,w
	movwf	d0
	movlw	0x1B
	bcf	STATUS,C
	rlf	d0,f
	btfsc	STATUS,C
	xorwf	d0,f

	movf	block+0xD,w		;
	xorwf	block+0xE,w
	xorwf	aux2,w
	movwf	d1
	movlw	0x1B
	bcf	STATUS,C
	rlf	d1,f
	btfsc	STATUS,C
	xorwf	d1,f

	
	movf	block+0xE,w		;
	xorwf	block+0xF,w
	xorwf	aux1,w
	movwf	d2
	movlw	0x1B
	bcf	STATUS,C
	rlf	d2,f
	btfsc	STATUS,C
	xorwf	d2,f


	movf	d0,w			;
	xorwf	aux3,w
	xorwf	block+0xC,f

	movf	d1,w			;
	xorwf	aux3,w
	xorwf	block+0xD,f
	
	movf	d2,w			;
	xorwf	aux3,w
	xorwf	block+0xE,f

	movf	block+0xC,w		;
	xorwf	block+0xD,w
	xorwf	block+0xE,w
	xorwf	aux,w
	movwf	block+0xF
	return

; ********************* ENC_KEY_SCHEDULE **********************
; This procedure takes the actual key and generates a new one.
; The equivalent C code fragment as designed by Paulo Barreto 
; and Vincent Rijmen: (key taken as a 4x4 matrix)
;
; 	word32 rcon[30] =      {0x01,0x02, 0x04, 0x08, 
;				0x10, 0x20, 0x40, 0x80, 
;				0x1b, 0x36, 0x6c, 0xd8, 
;				0xab, 0x4d, 0x9a, 0x2f, 
;				0x5e, 0xbc, 0x63, 0xc6, 
;				0x97, 0x35, 0x6a, 0xd4, 
;				0xb3, 0x7d, 0xfa, 0xef, 
;				0xc5, 0x91, };
;
;
;	rconpointer=0;
; 	for(i = 0; i < 4; i++)
; 	 key[i][0] ^= S[key[(i+1)%4][KC-1]];
;	key[0][0] ^= rcon[rconpointer++];
;       for(j = 1; j < KC; j++)
;	 for(i = 0; i < 4; i++) 
;	  key[i][j] ^= key[i][j-1];
;
;	
; Input: 	key
; Outup: 	
; Globals: 	aux,rcon
; ***************************************************************

enc_key_schedule:

	call	calc_s_table_based_values
	movf	rcon,w				; key[0] ^= rcon
	xorwf	key,f
	bcf	STATUS,C
	rlf	rcon,f				; rcon = xtime(rcon)
	btfss	STATUS,C
	goto	complete_round
	movlw	0x1B
	movwf	rcon
complete_round:
	movf	key+0x0,w		; This is equivalent to the 
	xorwf	key+0x4,f		; XOR of each column with the
					; previous one
	movf	key+0x1,w
	xorwf	key+0x5,f
					; first column1 ^= column0
	movf	key+0x2,w
	xorwf	key+0x6,f

	movf	key+0x3,w
	xorwf	key+0x7,f
					; column2 ^= column1
	movf	key+0x4,w
	xorwf	key+0x8,f

	movf	key+0x5,w
	xorwf	key+0x9,f

	movf	key+0x6,w
	xorwf	key+0xA,f

	movf	key+0x7,w
	xorwf	key+0xB,f
					; column3 ^= column2
	movf	key+0x8,w
	xorwf	key+0xC,f

	movf	key+0x9,w
	xorwf	key+0xD,f

	movf	key+0xA,w
	xorwf	key+0xE,f

	movf	key+0xB,w
	xorwf	key+0xF,f
	return


calc_s_table_based_values:

	movlw	HIGH S_table		; set the table segment
	movwf	PCLATH

	movf	key+0xD,w			
	btfss	STATUS,Z
	goto	put_S_table_0xD
	movlw	D'99'
	xorwf	key+0x0,f
	goto	read_key_0xE

put_S_table_0xD:
	decf	key+0xD,w		; key[0x0] ^= s_box[key[0xD]]
	call	S_table
	xorwf	key+0x0,f

read_key_0xE:

	movf	key+0xE,w			
	btfss	STATUS,Z
	goto	put_S_table_0xE
	movlw	D'99'
	xorwf	key+0x1,f
	goto	read_key_0xF

put_S_table_0xE:
	decf	key+0xE,w		; key[0x1] ^= s_box[key[0xE]]	
	call	S_table
	xorwf	key+0x1,f

read_key_0xF:

	movf	key+0xF,w			
	btfss	STATUS,Z
	goto	put_S_table_0xF
	movlw	D'99'
	xorwf	key+0x2,f
	goto	read_key_0xC

put_S_table_0xF:
	decf	key+0xF,w		; key[0x2] ^= s_box[key[0xF]]	
	call	S_table
	xorwf	key+0x2,f

read_key_0xC:

	movf	key+0xC,w			
	btfss	STATUS,Z
	goto	put_S_table_0xC
	movlw	D'99'
	xorwf	key+0x3,f
	return				;

put_S_table_0xC:
	decf	key+0xC,w		; key[0x3] ^= s_box[key[0xC]]	
	call	S_table
	xorwf	key+0x3,f
	return


; *********************** DEC_KEY_SCHEDULE **********************
; This procedure takes the actual key and generates a new one in
; the exact inverse sequence used by enc_key_schedule function
; 
;	
; Input: 	key
; Outup: 	
; Globals: 	aux,rcon
; ***************************************************************

dec_key_schedule:

;  
;first column3 ^= column2

	movf	key+0x8,w	
	xorwf	key+0xC,f

	movf	key+0x09,w	
	xorwf	key+0x0D,f

	movf	key+0x0A,w	
	xorwf	key+0x0E,f

	movf	key+0x0B,w	
	xorwf	key+0x0F,f

;column2 ^= column1

	movf	key+0x04,w	
	xorwf	key+0x08,f

	movf	key+0x05,w	
	xorwf	key+0x09,f

	movf	key+0x06,w	
	xorwf	key+0x0A,f

	movf	key+0x07,w	
	xorwf	key+0x0B,f

;first column1 ^= column0

	movf	key+0x00,w	
	xorwf	key+0x04,f

	movf	key+0x01,w	
	xorwf	key+0x05,f

	movf	key+0x02,w	
	xorwf	key+0x06,f

	movf	key+0x03,w	
	xorwf	key+0x07,f

	movf	rcon,w			; key[ox0] ^= rcon,  rcon>>=1, if carry rcon = 0x80
	xorwf	key,f			
	bcf	STATUS,C		; this is equivalent to the
	rrf	rcon,f			; read of rcon table

	btfss	STATUS,C		; 
	goto	complete_decrypt_round
	movlw	0x80
	movwf	rcon
complete_decrypt_round:
	goto	calc_s_table_based_values		


; ************************** DECRYPT ******************************
; This procedure decipher the plain text in block using key
; Inputs: 	The encrypted text in block[], 
;		The decipher_key in key[]; this key is NOT the 
;		original cipher key used to cipher the plain text, 
;		but the scheduled original.
; Outups: 	The plain text in block 
;		The original cipher key	in key[]
; OBS: 		This implementation follows as possible the original
;		structure and function names used by the reference
;		implementation of AES
; *****************************************************************

decrypt:
	movlw	0x36
	movwf	rcon			; RCON initialization
	call	key_addition		; initial key addition
	movlw	ROUNDS
	movwf	round_counter		;
	goto	dec_first_round
loop_decrypt:
	call	inv_mix_column
dec_first_round:
 	call	substitution_Si		; 
 	call	dec_shift_row
	call	dec_key_schedule
	call	key_addition		; 
	decfsz	round_counter,f
	goto	loop_decrypt
	return
	

; ***************** INIT_DECRYPTION_KEY ***************************
; This procedure creates the initial deciphering key from the initial 
; cyphering key by multiple application of enc_key_schedule
; Inputs: 	The plain_text in block[]
;		The cipher_key in key[]
; Outups: 	The ciphered text in block
;		The rounded (decipher) key in key[]
; OBS: 		This implementation follows as possible the original
;		structure and function names used by the reference
;		implementation of AES
; *****************************************************************


init_decryption_key:
	movlw	0x01
	movwf	rcon
	movlw	0xA
	movwf	round_counter
loop_init_decryption_key:
	call	enc_key_schedule
	decfsz	round_counter,f
	goto	loop_init_decryption_key
	return



	END	  
		

